@@ -17,6 +17,7 @@ module Database.Persist.TH
     , MkPersistSettings
     , mpsBackend
     , mpsGeneric
+    , mpsPrefixFields
     , mkPersistSettings
     , sqlSettings
     , sqlOnlySettings
@@ -120,7 +121,7 @@ getSqlType allEntities ent =
 
         mEmbedded (FTTypeCon Just{} _) = Nothing
         mEmbedded (FTTypeCon Nothing n) = let name = HaskellName n in
-            find ((name ==) . entityHaskell) allEntities 
+            find ((name ==) . entityHaskell) allEntities
         mEmbedded (FTList x) = mEmbedded x
         mEmbedded (FTApp x y) = maybe (mEmbedded y) Just (mEmbedded x)
 
@@ -174,6 +175,8 @@ data MkPersistSettings = MkPersistSettings
     -- ^ Create generic types that can be used with multiple backends. Good for
     -- reusable code, but makes error messages harder to understand. Default:
     -- True.
+    , mpsPrefixFields :: Bool
+    -- ^ Prefix field names with the model name. Default: True.
     }
 
 -- | Create an @MkPersistSettings@ with default values.
@@ -182,6 +185,7 @@ mkPersistSettings :: Type -- ^ Value for 'mpsBackend'
 mkPersistSettings t = MkPersistSettings
     { mpsBackend = t
     , mpsGeneric = True -- FIXME switch default to False in the future
+    , mpsPrefixFields = True
     }
 
 -- | Use the 'SqlPersist' backend.
@@ -194,8 +198,10 @@ sqlSettings = mkPersistSettings $ ConT ''SqlBackend
 sqlOnlySettings :: MkPersistSettings
 sqlOnlySettings = sqlSettings { mpsGeneric = False }
 
-recName :: Text -> Text -> Text
-recName dt f = lowerFirst dt ++ upperFirst f
+recName :: MkPersistSettings -> Text -> Text -> Text
+recName mps dt f
+  | mpsPrefixFields mps = lowerFirst dt ++ upperFirst f
+  | otherwise           = lowerFirst f
 
 lowerFirst :: Text -> Text
 lowerFirst t =
@@ -215,7 +221,7 @@ dataTypeDec mps t =
     $ map (mkName . unpack) $ entityDerives t
   where
     mkCol x FieldDef {..} =
-        (mkName $ unpack $ recName x $ unHaskellName fieldHaskell,
+        (mkName $ unpack $ recName mps x $ unHaskellName fieldHaskell,
          if fieldStrict then IsStrict else NotStrict,
          pairToType mps backend (fieldType, nullable fieldAttrs)
         )
@@ -232,12 +238,14 @@ dataTypeDec mps t =
         | otherwise = [RecC name cols]
 
     sumCon fd = NormalC
-        (sumConstrName t fd)
+        (sumConstrName mps t fd)
         [(NotStrict, pairToType mps backend (fieldType fd, NotNullable))]
 
-sumConstrName :: EntityDef a -> FieldDef b -> Name
-sumConstrName t FieldDef {..} = mkName $ unpack $ concat
-    [ unHaskellName $ entityHaskell t
+sumConstrName :: MkPersistSettings -> EntityDef a -> FieldDef b -> Name
+sumConstrName mps t FieldDef {..} = mkName $ unpack $ concat
+    [ if mpsPrefixFields mps
+        then unHaskellName $ entityHaskell t
+        else ""
     , upperFirst $ unHaskellName fieldHaskell
     , "Sum"
     ]
@@ -334,8 +342,8 @@ degen [] =
      in [Clause [WildP] (NormalB err) []]
 degen x = x
 
-mkToPersistFields :: String -> EntityDef a -> Q Dec
-mkToPersistFields constr ed@EntityDef { entitySum = isSum, entityFields = fields } = do
+mkToPersistFields :: MkPersistSettings -> String -> EntityDef a -> Q Dec
+mkToPersistFields mps constr ed@EntityDef { entitySum = isSum, entityFields = fields } = do
     clauses <-
         if isSum
             then sequence $ zipWith goSum fields [1..]
@@ -354,7 +362,7 @@ mkToPersistFields constr ed@EntityDef { entitySum = isSum, entityFields = fields
 
     goSum :: FieldDef a -> Int -> Q Clause
     goSum fd idx = do
-        let name = sumConstrName ed fd
+        let name = sumConstrName mps ed fd
         enull <- [|SomePersistField PersistNull|]
         let beforeCount = idx - 1
             afterCount = fieldCount - idx
@@ -425,8 +433,8 @@ isNotNull :: PersistValue -> Bool
 isNotNull PersistNull = False
 isNotNull _ = True
 
-mkFromPersistValues :: EntityDef a -> Q [Clause]
-mkFromPersistValues t@(EntityDef { entitySum = False }) = do
+mkFromPersistValues :: MkPersistSettings -> EntityDef a -> Q [Clause]
+mkFromPersistValues mps t@(EntityDef { entitySum = False }) = do
     nothing <- [|Left $(liftT $ "Invalid fromPersistValues input. Entity: " `mappend` entName)|]
     let cons' = ConE $ mkName $ unpack $ entName
     xs <- mapM (const $ newName "x") $ entityFields t
@@ -450,7 +458,7 @@ mkFromPersistValues t@(EntityDef { entitySum = False }) = do
     entName = unHaskellName $ entityHaskell t
     go ap' x y = InfixE (Just x) ap' (Just y)
 
-mkFromPersistValues t@(EntityDef { entitySum = True }) = do
+mkFromPersistValues mps t@(EntityDef { entitySum = True }) = do
     nothing <- [|Left $(liftT $ "Invalid fromPersistValues input: sum type with all nulls. Entity: " `mappend` entName)|]
     clauses <- mkClauses [] $ entityFields t
     return $ clauses `mappend` [Clause [WildP] (NormalB nothing) []]
@@ -465,7 +473,7 @@ mkFromPersistValues t@(EntityDef { entitySum = True }) = do
                 , [VarP x]
                 , map (const null') after
                 ]
-            constr = ConE $ sumConstrName t field
+            constr = ConE $ sumConstrName mps t field
         fmap' <- [|fmap|]
         fs <- [|fromPersistValue $(return $ VarE x)|]
         let guard' = NormalG $ VarE 'isNotNull `AppE` VarE x
@@ -478,8 +486,8 @@ type Lens s t a b = forall f. Functor f => (a -> f b) -> s -> f t
 lens :: (s -> a) -> (s -> b -> t) -> Lens s t a b
 lens sa sbt afb s = fmap (sbt s) (afb $ sa s)
 
-mkLensClauses :: EntityDef a -> Q [Clause]
-mkLensClauses t = do
+mkLensClauses :: MkPersistSettings -> EntityDef a -> Q [Clause]
+mkLensClauses mps t = do
     lens' <- [|lens|]
     getId <- [|entityKey|]
     setId <- [|\(Entity _ value) key -> Entity key value|]
@@ -497,11 +505,11 @@ mkLensClauses t = do
         else return $ idClause : map (toClause lens' getVal dot keyName valName xName) (entityFields t)
   where
     toClause lens' getVal dot keyName valName xName f = Clause
-        [ConP (mkName $ unpack $ unHaskellName (entityHaskell t) ++ upperFirst (unHaskellName $ fieldHaskell f)) []]
+        [ConP (filterConName mps t f) []]
         (NormalB $ lens' `AppE` getter `AppE` setter)
         []
       where
-        fieldName = mkName $ unpack $ recName (unHaskellName $ entityHaskell t) (unHaskellName $ fieldHaskell f)
+        fieldName = mkName $ unpack $ recName mps (unHaskellName $ entityHaskell t) (unHaskellName $ fieldHaskell f)
         getter = InfixE (Just $ VarE fieldName) dot (Just getVal)
         setter = LamE
             [ ConP 'Entity [VarP keyName, VarP valName]
@@ -512,7 +520,7 @@ mkLensClauses t = do
                 [(fieldName, VarE xName)]
 
     toSumClause lens' keyName valName xName f = Clause
-        [ConP (mkName $ unpack $ unHaskellName (entityHaskell t) ++ upperFirst (unHaskellName $ fieldHaskell f)) []]
+        [ConP (filterConName mps t f) []]
         (NormalB $ lens' `AppE` getter `AppE` setter)
         []
       where
@@ -520,7 +528,7 @@ mkLensClauses t = do
         getter = LamE
             [ ConP 'Entity [WildP, VarP valName]
             ] $ CaseE (VarE valName)
-            $ Match (ConP (sumConstrName t f) [VarP xName]) (NormalB $ VarE xName) []
+            $ Match (ConP (sumConstrName mps t f) [VarP xName]) (NormalB $ VarE xName) []
 
             -- FIXME It would be nice if the types expressed that the Field is
             -- a sum type and therefore could result in Maybe.
@@ -529,7 +537,7 @@ mkLensClauses t = do
             [ ConP 'Entity [VarP keyName, WildP]
             , VarP xName
             ]
-            $ ConE 'Entity `AppE` VarE keyName `AppE` (ConE (sumConstrName t f) `AppE` VarE xName)
+            $ ConE 'Entity `AppE` VarE keyName `AppE` (ConE (sumConstrName mps t f) `AppE` VarE xName)
 
 mkEntity :: MkPersistSettings -> EntityDef SqlType -> Q [Dec]
 mkEntity mps t = do
@@ -537,8 +545,8 @@ mkEntity mps t = do
     let nameT = unHaskellName $ entityHaskell t
     let nameS = unpack nameT
     let clazz = ConT ''PersistEntity `AppT` genericDataType mps (unHaskellName $ entityHaskell t) (VarT $ mkName "backend")
-    tpf <- mkToPersistFields nameS t
-    fpv <- mkFromPersistValues t
+    tpf <- mkToPersistFields mps nameS t
+    fpv <- mkFromPersistValues mps t
     utv <- mkUniqueToValues $ entityUniques t
     puk <- mkUniqueKeys t
 
@@ -560,7 +568,7 @@ mkEntity mps t = do
                     genericDataType mps nameT $ mpsBackend mps
             | otherwise = id
 
-    lensClauses <- mkLensClauses t
+    lensClauses <- mkLensClauses mps t
 
     return $ addSyn
       [ dataTypeDec mps t
@@ -585,8 +593,14 @@ mkEntity mps t = do
         , FunD 'persistFieldDef (map snd fields)
         , TySynInstD
             ''PersistEntityBackend
+#if MIN_VERSION_template_haskell(2,9,0)
+            (TySynEqn
+               [genericDataType mps (unHaskellName $ entityHaskell t) $ VarT $ mkName "backend"]
+               (backendDataType mps))
+#else
             [genericDataType mps (unHaskellName $ entityHaskell t) $ VarT $ mkName "backend"]
             (backendDataType mps)
+#endif
         , FunD 'persistIdField [Clause [] (NormalB $ ConE $ mkName $ unpack $ unHaskellName (entityHaskell t) ++ "Id") []]
         , FunD 'fieldLens lensClauses
         ]
@@ -628,15 +642,6 @@ persistFieldFromEntity mps e = do
     where
       entityName = (unpack $ unHaskellName $ entityHaskell e)
 
-updateConName :: Text -> Text -> PersistUpdate -> Text
-updateConName name s pu = concat
-    [ name
-    , upperFirst s
-    , case pu of
-        Assign -> ""
-        _ -> pack $ show pu
-    ]
-
 -- | Apply the given list of functions to the same @EntityDef@s.
 --
 -- This function is useful for cases such as:
@@ -669,7 +674,7 @@ mkDeleteCascade mps defs = do
   where
     getDeps :: EntityDef a -> [Dep]
     getDeps def =
-        concatMap getDeps' $ entityFields def
+        concatMap getDeps' $ entityFields $ fixEntityDef def
       where
         getDeps' :: FieldDef a -> [Dep]
         getDeps' FieldDef {..} =
@@ -696,13 +701,12 @@ mkDeleteCascade mps defs = do
             mkStmt dep = NoBindS
                 $ dcw `AppE`
                   ListE
-                    [ filt `AppE` ConE (mkName $ unpack filtName)
+                    [ filt `AppE` ConE filtName
                            `AppE` (left `AppE` val (depSourceNull dep))
                            `AppE` eq
                     ]
               where
-                filtName = unHaskellName (depSourceTable dep) ++
-                           upperFirst (unHaskellName $ depSourceField dep)
+                filtName = filterConName' mps (depSourceTable dep) (depSourceField dep)
                 val (Nullable ByMaybeAttr) = just `AppE` VarE key
                 val _                      =             VarE key
 
@@ -948,10 +952,7 @@ mkField mps et cd = do
                 []
     return (con, cla)
   where
-    name = mkName $ unpack $ concat
-        [ unHaskellName $ entityHaskell et
-        , upperFirst $ unHaskellName $ fieldHaskell cd
-        ]
+    name = filterConName mps et cd
     maybeTyp =
         if nullable (fieldAttrs cd) == Nullable ByMaybeAttr
             then ConT ''Maybe `AppT` typ
@@ -966,6 +967,23 @@ mkField mps et cd = do
                     `AppT` genericDataType mps ft (VarT $ mkName "backend")
             Nothing -> ftToType $ fieldType cd
 
+filterConName :: MkPersistSettings
+              -> EntityDef sqlType1
+              -> FieldDef sqlType2
+              -> Name
+filterConName mps entity field = filterConName' mps (entityHaskell entity) (fieldHaskell field)
+
+filterConName' :: MkPersistSettings
+               -> HaskellName -- ^ table
+               -> HaskellName -- ^ field
+               -> Name
+filterConName' mps entity field = mkName $ unpack $ concat
+    [ if mpsPrefixFields mps || field == HaskellName "Id"
+        then unHaskellName entity
+        else ""
+    , upperFirst $ unHaskellName field
+    ]
+
 ftToType :: FieldType -> Type
 ftToType (FTTypeCon Nothing t) = ConT $ mkName $ unpack t
 ftToType (FTTypeCon (Just m) t) = ConT $ mkName $ unpack $ concat [m, ".", t]
